package lu.tudor.santec.dicom.anonymizer;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.dcm4che2.data.DicomObject;
import org.dcm4che2.data.Tag;
import org.dcm4che2.data.VR;
import org.dcm4che2.io.DicomInputStream;
import org.dcm4che2.io.DicomOutputStream;

/**
 * Helper class that provides methods to anonymize dicom files
 * 
 *
 * @author Johannes Hermen johannes.hermen(at)tudor.lu
 *
 * @version
 * <br>$Log: DicomAnonymizer.java,v $
 * <br>Revision 1.4  2008-11-13 13:20:30  hermen
 * <br>added bugfixes from Marc Guenther
 * <br>
 * <br>Revision 1.3  2008-10-30 10:28:38  hermen
 * <br>redesign and adaption to dcm4che 2.0.15
 * <br>
 *
 */
public class DicomAnonymizer {

    	public static final int[] PATIENT_TAGS = {
	    Tag.PatientName,
	    Tag.PatientID,
	    Tag.PatientAddress,
	    Tag.PatientBirthDate,
	    Tag.PatientBirthName,
	    Tag.PatientBirthTime,
	    Tag.PatientInstitutionResidence,
	    Tag.PatientInsurancePlanCodeSequence,
	    Tag.PatientMotherBirthName,
	    Tag.PatientReligiousPreference,
	    Tag.PatientTelephoneNumbers
    	};
    
    	public static final int[] INSTITUTION_TAGS = { 
    	    Tag.InstitutionAddress,
	    Tag.InstitutionalDepartmentName, 
	    Tag.InstitutionCodeSequence,
	    Tag.ReferringPhysicianName, 
	    Tag.ReferringPhysicianAddress,
	    Tag.NameOfPhysiciansReadingStudy, 
	    Tag.PatientInstitutionResidence,
	    Tag.RequestingPhysician,
	    Tag.RequestingPhysicianIdentificationSequence 
	};
    	
    	
    	public static final int[] MANUFACTURER_TAGS = {
	    Tag.Manufacturer,
	    Tag.ManufacturerModelName
    	};
    
    
    
	public DicomAnonymizer() {
	    
	}
	
	/**
	 * Anonymizes a Dicom File by removing all 
	 * - Patient
	 * - Institution
	 * - Manufacturer 
	 * Information
	 * @param f
	 */
	public static void anonymizeFile(File f) throws Exception{
	    
	    DicomObject dObj = readDicomObject(f);
	    
	    removePatientInfo(dObj, "", "");
	    removeInstitutionInfo(dObj, "");
	    removeManufacturerInfo(dObj);
	    
	    writeDicomFile(dObj, f);
	}
	
	/**
	 * replaces the given dicom tag (if existing) with the given value.
	 * @param dObj
	 * @param tag
	 * @param newValue
	 */
	public static void replaceTag(DicomObject dObj, int tag, String newValue) {
	    if (tag != 0 && dObj.contains(tag)) {
		VR vr = dObj.vrOf(tag);
		dObj.putString(tag, vr, newValue);
	    }
	}
	
	/**
	 * adds/replaces the given dicom tag with the given value.
	 * @param dObj
	 * @param tag
	 * @param newValue
	 */
	public static void addTag(DicomObject dObj, int tag, String newValue) {
		VR vr = dObj.vrOf(tag);
		dObj.putString(tag, vr, newValue);
	}
	
	
	public static DicomObject removePatientInfo(DicomObject dObj, String patientID, String newName) {
	    
	    if (newName == null || "".equals(newName)) {
        	    newName = "ANONYMOUS";
        	    try {
        		String gender = dObj.getString(Tag.PatientSex);		
        		newName += "_" + gender;
        		Date bDate = dObj.getDate(Tag.PatientBirthDate);
        		Calendar cal = new GregorianCalendar();
        		cal.setTime(bDate);
        		newName += "_" + cal.get(Calendar.YEAR); 
        	    } catch (Exception e) {
        	    }
	    }
	    
	    for (int tag : PATIENT_TAGS) {
		replaceTag(dObj, tag, "");
	    }
	    
	    addTag(dObj, Tag.PatientName, newName);
	    addTag(dObj, Tag.PatientID, patientID);
	    addTag(dObj, Tag.PatientIdentityRemoved, "true");
	    
	    return dObj;
	}
	
	public static DicomObject removeInstitutionInfo(DicomObject dObj, String institutionName) {
	    
	    for (int tag : INSTITUTION_TAGS) {
		replaceTag(dObj, tag, "");
	    }
	    addTag(dObj, Tag.InstitutionName, institutionName);
	    
	    return dObj;
	}
	
	public static DicomObject removeManufacturerInfo(DicomObject dObj) {
	    	
	    for (int tag : MANUFACTURER_TAGS) {
		replaceTag(dObj, tag, "");
	    }
	    
	    return dObj;
	}
	
	/**
	 * Read the file into a dicom object
	 * @param f the dicom file
	 * @return the read dicom object
	 */
	public static DicomObject readDicomObject(File f) {
	    DicomObject dcmObj;
	    DicomInputStream din = null;
	    try {
	        din = new DicomInputStream(f);
	        dcmObj = din.readDicomObject();
	        return dcmObj;
	    }
	    catch (IOException e) {
		e.printStackTrace();
	        return null;
	    }
	    finally {
	        try {
	            din.close();
	        }
	        catch (IOException ignore) {
	        }
	    }
	}
	
	/**
	 * write the dicom object to a file
	 * @param dObj the object to write
	 * @param f the file to write to
	 */
	public static void writeDicomFile(DicomObject dObj, File f) {
	    FileOutputStream fos;
	    try {
	        fos = new FileOutputStream(f);
	    }
	    catch (FileNotFoundException e) {
	        e.printStackTrace();
	        return;
	    }
	    BufferedOutputStream bos = new BufferedOutputStream(fos);
	    DicomOutputStream dos = new DicomOutputStream(bos);
	    try {
	        dos.writeDicomFile(dObj);
	    }
	    catch (IOException e) {
	        e.printStackTrace();
	        return;
	    }
	    finally {
	        try {
	            dos.close();
	        }
	        catch (IOException ignore) {
	        }
	    }
	}
	
	public static void main(String[] args) {
		try {
		    DicomAnonymizer.anonymizeFile(new File("/home/hermenj/1"));
		} catch (Exception e) {
		    // TODO Auto-generated catch block
		    e.printStackTrace();
		}
	}
	
}

